__all__ = ()

from .line_info import get_file_line


EXPRESSION_LINE_COUNT_MAX = 21


ACTION_TYPE_NONE = 0
ACTION_TYPE_PARENTHESES_OPEN_BRACKET = 1
ACTION_TYPE_PARENTHESES_OPEN_BOX = 2
ACTION_TYPE_PARENTHESES_OPEN_CURVY = 3
ACTION_TYPE_PARENTHESES_CLOSE_BRACKET = 4
ACTION_TYPE_PARENTHESES_CLOSE_BOX = 5
ACTION_TYPE_PARENTHESES_CLOSE_CURVY = 6
ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE = 7
ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE = 8
ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE = 9
ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE = 10
ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE = 11
ACTION_TYPE_STRING_CLOSE_DOUBLE_QUOTE = 12
ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE = 13
ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE = 14
ACTION_TYPE_LINE_ESCAPE = 15
ACTION_FORWARD = 16
ACTION_BACKWARDS = 17


CHARACTER_TO_ACTION = {
    '(': ACTION_TYPE_PARENTHESES_OPEN_BRACKET,
    '{': ACTION_TYPE_PARENTHESES_OPEN_CURVY,
    '[': ACTION_TYPE_PARENTHESES_OPEN_BOX,
    ')': ACTION_TYPE_PARENTHESES_CLOSE_BRACKET,
    ']': ACTION_TYPE_PARENTHESES_CLOSE_BOX,
    '}': ACTION_TYPE_PARENTHESES_CLOSE_CURVY,
}


ACTION_TYPE_STRING_OPEN = {
    ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE,
    ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE,
    ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE,
    ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE,
}


DIRECTION_NONE = 0
DIRECTION_FORWARD = 1
DIRECTION_BACKWARD = 2
DIRECTION_REPEAT_LAST = 3


ACTION_DIRECTIONS = {
    ACTION_TYPE_NONE: DIRECTION_NONE,
    ACTION_TYPE_PARENTHESES_OPEN_BRACKET: DIRECTION_FORWARD,
    ACTION_TYPE_PARENTHESES_OPEN_BOX: DIRECTION_FORWARD,
    ACTION_TYPE_PARENTHESES_OPEN_CURVY: DIRECTION_FORWARD,
    ACTION_TYPE_PARENTHESES_CLOSE_BRACKET: DIRECTION_BACKWARD,
    ACTION_TYPE_PARENTHESES_CLOSE_BOX: DIRECTION_BACKWARD,
    ACTION_TYPE_PARENTHESES_CLOSE_CURVY: DIRECTION_BACKWARD,
    
    ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_CLOSE_DOUBLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE: DIRECTION_REPEAT_LAST,
    ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE: DIRECTION_REPEAT_LAST,
    
    ACTION_TYPE_LINE_ESCAPE: DIRECTION_FORWARD,
    ACTION_FORWARD: DIRECTION_FORWARD,
    ACTION_BACKWARDS: DIRECTION_BACKWARD,
}

STRING_OPEN_CLOSE_PAIRS = {
    (ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE, ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE),
    (ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE, ACTION_TYPE_STRING_CLOSE_DOUBLE_QUOTE),
    (ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE, ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE),
    (ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE, ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE),
    (ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE, ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE),
    (ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE, ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE),
}


PARENTHESES_OPEN_CLOSE_PAIRS = {
    (ACTION_TYPE_PARENTHESES_OPEN_BRACKET, ACTION_TYPE_PARENTHESES_CLOSE_BRACKET),
    (ACTION_TYPE_PARENTHESES_OPEN_BOX, ACTION_TYPE_PARENTHESES_CLOSE_BOX),
    (ACTION_TYPE_PARENTHESES_OPEN_CURVY, ACTION_TYPE_PARENTHESES_CLOSE_CURVY),
}


STRING_MUST_BE_PAIRED = {
    ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE,
    ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE,
    ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE,
    ACTION_TYPE_STRING_CLOSE_DOUBLE_QUOTE,
}

PARENTHESES_INVALID_PAIRS = {
    (ACTION_TYPE_PARENTHESES_OPEN_BRACKET, ACTION_TYPE_PARENTHESES_CLOSE_BOX),
    (ACTION_TYPE_PARENTHESES_OPEN_BRACKET, ACTION_TYPE_PARENTHESES_CLOSE_CURVY),
    (ACTION_TYPE_PARENTHESES_OPEN_BOX, ACTION_TYPE_PARENTHESES_CLOSE_BRACKET),
    (ACTION_TYPE_PARENTHESES_OPEN_BOX, ACTION_TYPE_PARENTHESES_CLOSE_CURVY),
    (ACTION_TYPE_PARENTHESES_OPEN_CURVY, ACTION_TYPE_PARENTHESES_CLOSE_BRACKET),
    (ACTION_TYPE_PARENTHESES_OPEN_CURVY, ACTION_TYPE_PARENTHESES_CLOSE_BOX),
}


def _parse_forward_next_action(line, index, length, source_action):
    """
    Parses the next action forward.
    
    Parameters
    ----------
    line : `str`
        The line to parse from.
    index : `int`
        The index to start parsing at.
    length : `int`
        Index to parse till.
    source_action : `int`
        The previous action.
    
    Returns
    -------
    index : `int`
    action : `int`
    """
    if source_action in ACTION_TYPE_STRING_OPEN:
        return _parse_forward_next_string(line, index, length, source_action)
    
    return _parse_forward_next_any(line, index, length)


def _parse_forward_next_string(line, index, length, source_action):
    """
    Parses till the string is ended.
    
    Parameters
    ----------
    line : `str`
        The line to parse from.
    index : `int`
        The index to start parsing at.
    length : `int`
        Index to parse till.
    
    Returns
    -------
    index : `int`
    action : `int`
    """
    if source_action == ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE:
        end_character = '\''
        triple_quited = False
        closer_action = ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE
    elif source_action == ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE:
        end_character = '"'
        triple_quited = False
        closer_action = ACTION_TYPE_STRING_CLOSE_DOUBLE_QUOTE
    elif source_action == ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE:
        end_character = '\''
        triple_quited = True
        closer_action = ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE
    elif source_action == ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE:
        end_character = '"'
        triple_quited = True
        closer_action = ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE
    else:
        end_character = '\''
        triple_quited = False
        closer_action = ACTION_TYPE_STRING_CLOSE_SINGLE_QUOTE
    
    
    while True:
        if index >= length:
            action = ACTION_TYPE_NONE
            break
        
        character = line[index]
        if character == '\\':
            index += 2
            continue
        
        if character == end_character:
            if triple_quited:
                if (
                    (index + 2 < length) and
                    (line[index + 1] == end_character) and
                    (line[index + 2] == end_character)
                ):
                    index += 2
                else:
                    index += 1
                    continue
            
            index += 1
            action = closer_action
            break
        
        index += 1
        continue
    
    return index, action


def _parse_forward_next_any(line, index, length):
    """
    Parses one forward till hits any action.
    
    Parameters
    ----------
    line : `str`
        The line to parse from.
    index : `int`
        The index to start parsing at.
    length : `int`
        Index to parse till.
    
    Returns
    -------
    index : `int`
    action : `int`
    """
    while True:
        if index >= length:
            action = ACTION_TYPE_NONE
            break
        
        character = line[index]
        
        if character == '#':
            action = ACTION_TYPE_NONE
            index = length
            break
        
        if character == '\\':
            action = ACTION_TYPE_LINE_ESCAPE
            index = length
            break
        
        try:
            action = CHARACTER_TO_ACTION[character]
        except KeyError:
            pass
        else:
            index += 1
            break
        
        if character in ('\'', '"'):
            if (
                (index + 2 < length) and
                (line[index + 1] == character) and
                (line[index + 2] == character)
            ):
                index += 2
                
                if character == '\'':
                    action = ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE
                else:
                    action = ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE
                    
            else:
                if character == '\'':
                    action = ACTION_TYPE_STRING_OPEN_SINGLE_QUOTE
                else:
                    action = ACTION_TYPE_STRING_OPEN_DOUBLE_QUOTE
            
            index += 1
            break
  
        index += 1
        continue
    
    return index, action


def _try_parse_triple_quote_from_behind(line, source_action):
    """
    Tries to parse a triple quoted string from behind returning the where it started and whether it was closed.
    
    Parameters
    ----------
    line : `str`
        Line to parse.
    source_action : `int`
        The last action executed.
    
    Returns
    -------
    length : `int`
    action : `int`
    """
    if source_action == ACTION_TYPE_STRING_OPEN_TRIPLE_SINGLE_QUOTE:
        end_character = '\''
        closer_action = ACTION_TYPE_STRING_CLOSE_TRIPLE_SINGLE_QUOTE
    elif source_action == ACTION_TYPE_STRING_OPEN_TRIPLE_DOUBLE_QUOTE:
        end_character = '"'
        closer_action = ACTION_TYPE_STRING_CLOSE_TRIPLE_DOUBLE_QUOTE
    else:
        return len(line), ACTION_TYPE_NONE 
    
    for index in range(2, len(line)):
        if line[index] != end_character:
            continue
        
        if line[index - 1] != end_character:
            continue
        
        if line[index - 2] != end_character:
            continue
        
        if index - 3 >= 0 and line[index - 3] == '\\':
            continue
        
        return index - 2, closer_action
    
    return 0, ACTION_TYPE_NONE


def _normalize_and_dedent(lines):
    """
    Normalises the lines and dedents them.
    
    Parameters
    ----------
    lines : `list<str>`
        The lines to work on.
    """
    if not lines:
        return
    
    for index in range(len(lines)):
        lines[index] = lines[index].rstrip()
    
    indentation_to_remove = -1
    for line in lines:
        if not line:
            continue
        
        count = 0
        for character in line:
            if character not in (' ', '\t'):
                break
            
            count += 1
            continue
        
        if indentation_to_remove == -1:
            indentation_to_remove = count
            continue
        
        if count < indentation_to_remove:
            indentation_to_remove = count
        continue
    
    
    if indentation_to_remove != -1:
        for index in range(len(lines)):
            lines[index] = lines[index][indentation_to_remove:]


def _is_syntax_valid(action_stack):
    """
    Returns whether the syntax looks valid.
    
    Parameters
    ----------
    action_stack : `list<int>`
        Action stack to read the syntax from.
    
    Returns
    -------
    is_syntax_valid : `bool`
    """
    if not action_stack:
        return True
    
    for action in action_stack:
        if action in STRING_MUST_BE_PAIRED:
            return False
    
    last_action = action_stack[0]
    for index in range(1, len(action_stack)):
        action = action_stack[index]
        if (last_action, action) in PARENTHESES_INVALID_PAIRS:
            return False
        
        last_action = action
        continue
    
    return True


def _get_action_direction(action_stack, last_direction):
    """
    Gets the direction for the next action.
    
    Parameters
    ----------
    action_stack : `list<int>`
        Action stack to pull the last action from.
    last_direction : `int`
        The last direction.
    
    Returns
    -------
    new_direction : `int`
    """
    if last_direction == DIRECTION_FORWARD:
        action = action_stack[-1]
    else:
        action = action_stack[0]
    
    new_direction = ACTION_DIRECTIONS.get(action, DIRECTION_NONE)
    if new_direction == DIRECTION_REPEAT_LAST:
        new_direction = last_direction
    
    return new_direction


def _pop_forward_actions(action_stack):
    """
    Pops every forward action.
    
    Parameters
    ----------
    action_stack : `list<int>`
        Action stack to pop from.
    """
    for index in reversed(range(len(action_stack))):
        if ACTION_DIRECTIONS.get(action_stack[index], DIRECTION_NONE) == DIRECTION_FORWARD:
            del action_stack[index - 1:]
            break


def _pop_backward_actions(action_stack):
    """
    Pops every backward action.
    
    Parameters
    ----------
    action_stack : `list<int>`
        Action stack to pop from.
    """
    for index in range(len(action_stack)):
        if ACTION_DIRECTIONS.get(action_stack[index], DIRECTION_NONE) != DIRECTION_BACKWARD:
            del action_stack[: index]
            break


def _merge_actions(action_stack):
    """
    Merges actions as applicable. Also returns the dummy ones.
    
    Parameters
    ----------
    action_stack : `list<int>`
        The action stack to operate on.
    """
    # remove in-middle escapes
    for index in reversed(range(len(action_stack) - 1)):
        if action_stack[index] == ACTION_TYPE_LINE_ESCAPE:
            del action_stack[index]
    
    # remove nulled actions
    for index in reversed(range(len(action_stack))):
        if action_stack[index] in (ACTION_TYPE_NONE, ACTION_FORWARD, ACTION_BACKWARDS):
            del action_stack[index]
    
    if len(action_stack) >= 2:
        # remove continuous strings.
        last_action = action_stack[-1]
        for index in reversed(range(len(action_stack) - 1)):
            action = action_stack[index]
            if (action, last_action) not in STRING_OPEN_CLOSE_PAIRS:
                last_action = action
                continue
            
            del action_stack[index: index + 2]
            last_action = ACTION_TYPE_NONE
            continue
    
    
    # Yeet parsed parentheses repeatedly
    while True:
        if len(action_stack) < 2:
            break
        
        hit_any = False
        
        last_action = action_stack[-1]
        for index in reversed(range(len(action_stack) - 1)):
            action = action_stack[index]
            if (action, last_action) not in PARENTHESES_OPEN_CLOSE_PAIRS:
                last_action = action
                continue
            
            del action_stack[index : index + 2]
            last_action = ACTION_TYPE_NONE
            hit_any = True
            continue
        
        if not hit_any:
            break


def get_expression_lines(file_name, line_index):
    """
    Tries the expression's lines at the given position.
    
    Parameters
    ----------
    file_name : `str`
        The file's name to get the expressions for.
    line_index : `int`
        The line's index in the file.
    
    Returns
    -------
    lines : `list<str>`
    shift : `int`
    syntax_valid : `bool`
    """
    action_stack = [ACTION_FORWARD]
    lines = []
    
    forward_line_index = line_index - 1
    backwards_line_index = line_index
    
    direction = DIRECTION_FORWARD
    syntax_valid = True
    
    while True:
        if not action_stack:
            break
        
        direction = _get_action_direction(action_stack, direction)
        if direction == DIRECTION_NONE:
            break
        
        if direction == DIRECTION_FORWARD:
            hit, line = get_file_line(file_name, forward_line_index + 1)
            if hit:
                forward_line_index += 1
            else:
                _pop_forward_actions(action_stack)
                direction = DIRECTION_BACKWARD
                continue
        else:
            hit, line = get_file_line(file_name, backwards_line_index - 1)
            if hit:
                backwards_line_index -= 1
            else:
                _pop_backward_actions(action_stack)
                direction = DIRECTION_FORWARD
                continue
        
        # Append line
        if direction == DIRECTION_FORWARD:
            lines.append(line)
        else:
            lines.insert(0, line)
        
        # Check max allowed length
        if len(lines) > EXPRESSION_LINE_COUNT_MAX:
            break
        
        # Get last action
        line_action_stack = []
        if direction == DIRECTION_FORWARD:
            action = action_stack[-1]
        else:
            action = action_stack[0]
        
        # Continue
        if DIRECTION_FORWARD:
            length = len(line)
            character_index = 0
            
            while character_index < length:
                character_index, action = _parse_forward_next_action(line, character_index, length, action)
                line_action_stack.append(action)
        else:
            length, action = _try_parse_triple_quote_from_behind(line, action)
            action_stack.insert(0, action)
            action = ACTION_TYPE_NONE
            
            character_index = 0
            
            while character_index < length:
                character_index, action = _parse_forward_next_action(line, character_index, length, action)
                line_action_stack.append(action)
        
        # Add newly parsed actions
        if direction == DIRECTION_FORWARD:
            action_stack.extend(line_action_stack)
        else:
            action_stack[:0] = line_action_stack
        
        _merge_actions(action_stack)
        syntax_valid = _is_syntax_valid(action_stack)
        if not syntax_valid:
            break
        
        continue
    
    _normalize_and_dedent(lines)
    return lines, line_index - backwards_line_index, syntax_valid



__all__ = ()

from reprlib import repr as short_repr

from ..utils import RichAttributeErrorBaseType, copy_docs, export

from .headers import CONNECTION, CONTENT_ENCODING, KEEP_ALIVE, TRANSFER_ENCODING, UPGRADE
from .helpers import HttpVersion11
from .keep_alive_info import KeepAliveInfo


CACHE_FLAG_UPGRADE_SET = 1 << 0
CACHE_FLAG_UPGRADE_VALUE = 1 << 1

CACHE_FLAG_CHUNKED_SET = 1 << 2
CACHE_FLAG_CHUNKED_VALUE = 1 << 3

CACHE_FLAG_ENCODING_SET = 1 << 4

CACHE_FLAG_KEEP_ALIVE_SET = 1 << 5


class RawMessage(RichAttributeErrorBaseType):
    """
    Base type of ``RawResponseMessage`` and ``RawRequestMessage``.
    
    Attributes
    ----------
    _cache_encoding : `None | str`
        Cache field for ``.encoding`` property.
    
    _cache_flags : `int`
        Cache flags for properties.
    
    _cache_keep_alive : `None | KeepAliveInfo`.
        cache field for ``.keep_alive`` property.
    
    headers : `IgnoreCaseMultiValueDictionary<str, str>`
        The headers of the http message.
    
    version : ``HttpVersion``
        The http version of the response.
    """
    __slots__ = ('_cache_encoding', '_cache_flags', '_cache_keep_alive', 'headers', 'version')
    
    def __new__(cls, version, headers):
        """
        Creates a new raw http message.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        
        headers : `IgnoreCaseMultiValueDictionary<str, str>`
            The headers of the http message.
        """
        self = object.__new__(cls)
        self._cache_encoding = None
        self._cache_flags = 0
        self._cache_keep_alive = None
        self.headers = headers
        self.version = version
        return self
    
    
    def __repr__(self):
        """Returns the http messages representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' version = ')
        repr_parts.append(repr(self.version))
        
        self._put_repr_parts_into(repr_parts)
        
        repr_parts.append(', headers = ')
        repr_parts.append(repr(self.headers))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Appends the embed field's representation parts.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            Representation parts to extend.
        """
        upgraded = self.upgraded
        if upgraded:
            repr_parts.append(', upgraded = ')
            repr_parts.append(repr(upgraded))
        
        chunked = self.chunked
        if chunked:
            repr_parts.append(', chunked = ')
            repr_parts.append(repr(chunked))
        
        keep_alive = self.keep_alive
        if (keep_alive is not None):
            repr_parts.append(', keep_alive = ')
            repr_parts.append(repr(keep_alive))
        
        encoding = self.encoding
        if (encoding is not None):
            repr_parts.append(', encoding = ')
            repr_parts.append(repr(encoding))
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be the same type as well.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        if self.headers != other.headers:
            return False
        
        if self.version != other.version:
            return False
        
        return True
    
    
    def __eq__(self, other):
        """Returns whether the two messages are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    @property
    def upgraded(self):
        """
        A get-set descriptor to access whether the message is upgraded.
        
        Returns
        -------
        upgrade : `bool`
        """
        cache_flags = self._cache_flags
        if cache_flags & CACHE_FLAG_UPGRADE_SET:
            return True if cache_flags & CACHE_FLAG_UPGRADE_VALUE else False
        
        headers = self.headers
        # Also require the `UPGRADE` header to be non-empty.
        if not headers.get(UPGRADE, ''):
            value = False
        
        else:
            try:
                connection = headers[CONNECTION]
            except KeyError:
                value = False
            else:
                value = (connection.casefold() == 'upgrade')
                if value:
                    cache_flags |= CACHE_FLAG_UPGRADE_VALUE
                else:
                    value = False
        
        cache_flags |= CACHE_FLAG_UPGRADE_SET
        self._cache_flags = cache_flags
        
        return value
    
    
    @upgraded.setter
    def upgraded(self, value):
        cache_flags = self._cache_flags
        cache_flags |= CACHE_FLAG_UPGRADE_SET
        
        if value:
            cache_flags |= CACHE_FLAG_UPGRADE_VALUE
        else:
            cache_flags &= ~CACHE_FLAG_UPGRADE_VALUE
        
        self._cache_flags = cache_flags
    
    
    @property
    def chunked(self):
        """
        Returns whether the respective http message' body is chunked.
        
        Returns
        -------
        chunked : `bool`
        """
        cache_flags = self._cache_flags
        if cache_flags & CACHE_FLAG_CHUNKED_SET:
            return True if cache_flags & CACHE_FLAG_CHUNKED_VALUE else False
        
        try:
            transfer_encoding = self.headers[TRANSFER_ENCODING]
        except KeyError:
            value = False
        else:
            value = ('chunked' in transfer_encoding.casefold())
            if value:
                cache_flags |= CACHE_FLAG_CHUNKED_VALUE
        
        cache_flags |= CACHE_FLAG_CHUNKED_SET
        self._cache_flags = cache_flags
        
        return value
    
    
    @chunked.setter
    def chunked(self, value):
        cache_flags = self._cache_flags
        cache_flags |= CACHE_FLAG_CHUNKED_SET
        
        if value:
            cache_flags |= CACHE_FLAG_CHUNKED_VALUE
        else:
            cache_flags &= ~CACHE_FLAG_CHUNKED_VALUE
        
        self._cache_flags = cache_flags
    
    
    @property
    def encoding(self):
        """
        Returns the body encoding set in the http message's header.
        
        Returns
        -------
        encoding : `None | str`
            If no encoding is set, defaults to `None`.
        """
        cache_flags = self._cache_flags
        if cache_flags & CACHE_FLAG_ENCODING_SET:
            return self._cache_encoding
        
        try:
            encoding = self.headers[CONTENT_ENCODING]
        except KeyError:
            encoding = None
        else:
            encoding = encoding.casefold()
        
        self._cache_flags = cache_flags | CACHE_FLAG_ENCODING_SET
        self._cache_encoding = encoding
        return encoding

        
    @encoding.setter
    def encoding(self, value):
        self._cache_flags |= CACHE_FLAG_ENCODING_SET
        self._cache_encoding = value
    
    
    @property
    def keep_alive(self):
        """
        Returns whether the respective http connection is keep-alive.
        
        Returns
        -------
        keep_alive : `bool`
        """
        cache_flags = self._cache_flags
        if cache_flags & CACHE_FLAG_KEEP_ALIVE_SET:
            return self._cache_keep_alive
        
        try:
            connection = self.headers[CONNECTION]
        except KeyError:
            keep_alive_enabled = self.version >= HttpVersion11
        else:
            keep_alive_enabled = ('close' not in connection.casefold())
        
        if keep_alive_enabled:
            keep_alive = KeepAliveInfo.from_header_value(self.headers.get(KEEP_ALIVE))
        else:
            keep_alive = None
        
        self._cache_flags = cache_flags | CACHE_FLAG_KEEP_ALIVE_SET
        self._cache_keep_alive = keep_alive
        
        return keep_alive
    
    
    @keep_alive.setter
    def keep_alive(self, value):
        self._cache_flags |= CACHE_FLAG_KEEP_ALIVE_SET
        self._cache_keep_alive = value


@export
class RawResponseMessage(RawMessage):
    """
    Represents a raw http response message.
    
    Attributes
    ----------
    _cache_encoding : `None | str`
        Cache field for ``.encoding`` property.
    
    _cache_flags : `int`
        Cache flags for properties.
    
    _cache_keep_alive : `None | KeepAliveInfo`.
        cache field for ``.keep_alive`` property.
    
    headers : `IgnoreCaseMultiValueDictionary<str, str>`
        The headers of the http message.
    
    reason : `None | str`
        Reason included with the response.
    
    status : `int`
        The response's status.
    
    version : ``HttpVersion``
        The http version of the response.
    """
    __slots__ = ('status', 'reason',)
    
    def __new__(cls, version, status, reason, headers):
        """
        Creates a new raw response message with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        
        status : `int`
            The response's status.
        
        reason : `None | str`
            Reason included with the response.
        
        headers : `IgnoreCaseMultiValueDictionary<str, str>`
            The headers of the http message.
        """
        self = RawMessage.__new__(cls, version, headers)
        
        self.reason = reason
        self.status = status
        
        return self


    @copy_docs(RawMessage._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(' status = ')
        repr_parts.append(repr(self.status))
        
        reason = self.reason
        if reason is not None:
            repr_parts.append(', reason = ')
            repr_parts.append(short_repr(reason))
        
        RawMessage._put_repr_parts_into(self, repr_parts)
    
    
    @copy_docs(RawMessage._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not RawMessage._is_equal_same_type(self, other):
            return False
        
        if self.reason != other.reason:
            return False
        
        if self.status != other.status:
            return False
        
        return True


@export
class RawRequestMessage(RawMessage):
    """
    Represents a raw http request message.
    
    Attributes
    ----------
    _cache_encoding : `None | str`
        Cache field for ``.encoding`` property.
    
    _cache_flags : `int`
        Cache flags for properties.
    
    _cache_keep_alive : `None | KeepAliveInfo`.
        cache field for ``.keep_alive`` property.
    
    headers : `IgnoreCaseMultiValueDictionary<str, str>`
        The headers of the http message.
    
    method : `str`
        The request's method.
    
    path : `str`
        The requested path.
    
    version : ``HttpVersion``
        The http version of the response.
    """
    __slots__ = ('method', 'path')
    
    def __new__(cls, version, method, path, headers):
        """
        Creates a new ``RawRequestMessage`` with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        
        method : `str`
            The request's method.
        
        path : `str`
            The requested path.
        
        headers : `IgnoreCaseMultiValueDictionary<str, str>`
            The headers of the http message.
        """
        self = RawMessage.__new__(cls, version, headers)
        
        self.method = method
        self.path = path
        
        return self
    
    
    @copy_docs(RawMessage._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        
        repr_parts.append(', method = ')
        repr_parts.append(repr(self.method))
        
        path = self.path
        if (path is not None):
            repr_parts.append(', path = ')
            repr_parts.append(repr(path))
        
        RawMessage._put_repr_parts_into(self, repr_parts)
    
    
    @copy_docs(RawMessage._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not RawMessage._is_equal_same_type(self, other):
            return False
        
        if self.method != other.method:
            return False
        
        if self.path != other.path:
            return False
        
        return True



__all__ = ()


def get_repeat_with_strategy_top(items):
    """
    Gets the top-most repeat on the given items.
    
    Parameters
    ----------
    items : `None | list`
        Items to get repeated part of.
    
    Returns
    -------
    repeat : `None | (int, int, int)`
        On detected repeat returns a tuple of `3` elements:
        - `start_shift` How much is the first element shifted
        - `chunk_size` How big 1 repeated chunk is.
        - `repeat` How much times was the chunk repeated.
    """
    if items is None:
        return
    
    items_length = len(items)
    for chunk_size in range(1, (items_length >> 1) + 1):
        for start_shift in range(0, items_length - chunk_size - 1):
            repeat = 0
            for additional_element_index in range(start_shift + chunk_size, items_length - chunk_size + 1, chunk_size):
                for shift in range(0, chunk_size):
                    if not (items[start_shift + shift] % items[additional_element_index + shift]):
                        break
                else:
                    repeat += 1
                    continue
                break
            
            if repeat * chunk_size >= 2:
                return start_shift, chunk_size, repeat + 1


def get_repeat_with_strategy_bot(items):
    """
    Gets the bot-most repeat on the given items.
    
    Parameters
    ----------
    items : `None | list`
        Items to get repeated part of.
    
    Returns
    -------
    repeat : `None | (int, int, int)`
        On detected repeat returns a tuple of `3` elements:
        - `start_shift` How much is the first element shifted
        - `chunk_size` How big 1 repeated chunk is.
        - `repeat` How much times was the chunk repeated.
    """
    if items is None:
        return
    
    items_length = len(items)
    for chunk_size in range(1, (items_length >> 1) + 1):
        for start_shift in range(items_length - chunk_size, chunk_size - 1, -1):
            repeat = 0
            for additional_element_index in range(start_shift - chunk_size, -1, -chunk_size):
                for shift in range(0, chunk_size):
                    if not (items[start_shift + shift] % items[additional_element_index + shift]):
                        break
                else:
                    repeat += 1
                    continue
                break
            
            if repeat * chunk_size >= 2:
                return start_shift - (chunk_size * repeat), chunk_size, repeat + 1



from .suggestion import *

from .attribute_error_helpers import *
from .exception_helpers import *
from .exception_representation_attribute_error import *
from .exception_representation_base import *
from .exception_representation_generic import *
from .exception_representation_syntax_error import *
from .representation_helpers import *
from .selection import *
from .syntax_error_helpers import *


# Do not pass the references down.
__all__ = ()



__all__ = ('HttpVersion', 'HttpVersion10', 'HttpVersion11', 'freeze_headers',)

from re import I as re_ignore_case, compile as re_compile
import socket as module_socket
from collections import namedtuple

from ..core import CancelledError, Task
from ..utils import IgnoreCaseMultiValueDictionary, include

from .exceptions import PayloadError


RawRequestMessage = include('RawRequestMessage')
RawResponseMessage = include('RawResponseMessage')


HttpVersion = namedtuple('HttpVersion', ('major', 'minor'))
HttpVersion10 = HttpVersion(1, 0)
HttpVersion11 = HttpVersion(1, 1)


BASIC_AUTH_DEFAULT_ENCODING = 'latin1'


_ipv4_pattern = '^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
_ipv6_pattern = (
    '^(?:(?:(?:[A-F0-9]{1,4}:){6}|(?=(?:[A-F0-9]{0,4}:){0,6}'
    '(?:[0-9]{1,3}\\.){3}[0-9]{1,3}$)(([0-9A-F]{1,4}:){0,5}|:)'
    '((:[0-9A-F]{1,4}){1,5}:|:)|::(?:[A-F0-9]{1,4}:){5})'
    '(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\\.){3}'
    '(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])|(?:[A-F0-9]{1,4}:){7}'
    '[A-F0-9]{1,4}|(?=(?:[A-F0-9]{0,4}:){0,7}[A-F0-9]{0,4}$)'
    '(([0-9A-F]{1,4}:){1,7}|:)((:[0-9A-F]{1,4}){1,7}|:)|(?:[A-F0-9]{1,4}:){7}'
    ':|:(:[A-F0-9]{1,4}){7})$'
)

_ipv4_regex = re_compile(_ipv4_pattern)
_ipv6_regex = re_compile(_ipv6_pattern, flags = re_ignore_case)
_ipv4_regex_b = re_compile(_ipv4_pattern.encode('utf-8'))
_ipv6_regex_b = re_compile(_ipv6_pattern.encode('utf-8'), flags = re_ignore_case)

del _ipv4_pattern, _ipv6_pattern


def is_ip_address(host):
    """
    Returns whether the given `host` is an ip address.
    
    Parameters
    ----------
    host : None | str | bytes-like`
        Host value.
    
    Returns
    -------
    is_ip_address : `bool`
    
    Raises
    ------
    TypeError
        - If `host`'s type is invalid.
    """
    if host is None:
        return False
    
    if isinstance(host, str):
        if _ipv4_regex.fullmatch(host) is not None:
            return True
        
        if _ipv6_regex.fullmatch(host) is not None:
            return True
        
        return False
        
    if isinstance(host, (bytes, bytearray, memoryview)):
        if _ipv4_regex_b.fullmatch(host) is not None:
            return True
        
        if _ipv6_regex_b.fullmatch(host) is not None:
            return True
        
        return False
    
    raise TypeError(
        f'`host` can be `None`, `str`, `bytes-like`, got {type(host).__name__}; {host!r}.'
    )


def is_ipv4_address(host):
    """
    Returns whether the given host is an `ipv4` ip address.
    
    Parameters
    ----------
    host : None | str | bytes-like`
        Host value.
    
    Returns
    -------
    is_ipv4_address : `bool`
    
    Raises
    ------
    TypeError
        - If `host`'s type is invalid.
    """
    if host is None:
        return False
    
    if isinstance(host, str):
        if _ipv4_regex.fullmatch(host) is not None:
            return True
        
        return False
    
    if isinstance(host, (bytes, bytearray, memoryview)):
        if _ipv4_regex_b.fullmatch(host) is not None:
            return True
        
        return False
    
    raise TypeError(
        f'`host` can be `None`, `str`, `bytes-like`, got {type(host).__name__}; {host!r}.'
    )


def is_ipv6_address(host):
    """
    Returns whether the given host is an `ipv6` ip address.
    
    Parameters
    ----------
    host : None | str | bytes-like`
        Host value.
    
    Returns
    -------
    is_ipv6_address : `bool`
    
    Raises
    ------
    TypeError
        - If `host`'s type is invalid.
    """
    if host is None:
        return False
    
    if isinstance(host, str):
        if _ipv6_regex.fullmatch(host) is not None:
            return True
        
        return False
    
    if isinstance(host, (bytes, bytearray, memoryview)):
        if _ipv6_regex_b.fullmatch(host) is not None:
            return True
        
        return False
    
    raise TypeError(
        f'`host` can be `None`, `str`, `bytes-like`, got {type(host).__name__}; {host!r}.'
    )


TIMEOUT_STATE_NONE = 0
TIMEOUT_STATE_TIMED_OUT = 1
TIMEOUT_STATE_CANCELLED = 2
TIMEOUT_STATE_EXITED = 3

class Timeout:
    """
    Implements async timeout feature as a context manager.
    
    Attributes
    ----------
    _handle : `None`, ``TimerHandle``
        Timer handle to cancel the respective task when it occurs. After timeout occurs, or the timeouter is
        cancelled, set as `None`.
    _loop : ``EventThread``
        The event loop to what the timeouter is bound to.
    _task : `None`, ``Task``
        The respective task what will be cancelled. Set as `None` at creation and when cancelled or exited.
    _state : `int`
        The timeouter's state.
        
        Can be one of the following:
        +---------------------------+-------+
        | Respective name           | Value |
        +===========================+=======+
        | TIMEOUT_STATE_NONE        | 0     |
        +---------------------------+-------+
        | TIMEOUT_STATE_TIMED_OUT   | 1     |
        +---------------------------+-------+
        | TIMEOUT_STATE_CANCELLED   | 2     |
        +---------------------------+-------+
        | TIMEOUT_STATE_EXITED      | 3     |
        +---------------------------+-------+
    """
    __slots__ = ('_handle', '_loop', '_task', '_state')
    
    def __new__(cls, loop, timeout):
        """
        Creates a new timeouter instance bound to the given loop.
        
        The timeout starts when the timeouter is created.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the timeouter will be bound to.
        timeout : `float`
            Time in seconds after the task is cancelled. When the cancellation reaches the context manager, raises
            `TimeoutError` instead.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._handle = loop.call_after(timeout, cls._timeout, self)
        self._task = None
        self._state = TIMEOUT_STATE_NONE
        return self
    
    
    def cancel(self):
        """
        Cancels the timeouter.
        
        If already cancelled, does nothing.
        """
        if self._state == TIMEOUT_STATE_CANCELLED:
            return
        
        self._state = TIMEOUT_STATE_CANCELLED
        
        handle = self._handle
        if (handle is not None):
            self._handle = None
            handle.cancel()
        
        self._task = None
    
    
    def __enter__(self):
        """
        Enters the timeouter as a context manager.
        
        Returns
        -------
        self : ``Timeout``
        
        Raises
        ------
        RuntimeError
            ``Timeout`` entered outside of a ``Task``.
        TimeoutError
            Timeout already occurred.
        """
        task = self._loop.current_task
        if (task is None):
            raise RuntimeError(
                f'`{type(self).__name__}` entered outside of a `{Task.__name__}`!'
            )
        
        state = self._state
        if state == TIMEOUT_STATE_NONE:
            self._task = task
        elif state == TIMEOUT_STATE_TIMED_OUT:
            raise TimeoutError
        elif state == TIMEOUT_STATE_CANCELLED:
            pass
        else:
            raise RuntimeError(
                f'`{type(self).__name__}` already used.'
            )
        
        return self
    
    
    def _timeout(self):
        """
        Timeouts the timeouter if not yet timed out nor cancelled.
        """
        if self._state != TIMEOUT_STATE_NONE:
            return
        
        self._state = TIMEOUT_STATE_TIMED_OUT
        
        handle = self._handle
        if (handle is not None):
            self._handle = None
        
        task = self._task
        if (task is not None):
            task.cancel()
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """
        Exits from the timeouter. If the timeout occurs, then raises ``TimeoutError`` from the received cancellation.
        """
        handle = self._handle
        if (handle is not None):
            self._handle = None
            handle.cancel()
        
        self._task = None
        
        state = self._state
        self._state = TIMEOUT_STATE_EXITED
        if (state == TIMEOUT_STATE_TIMED_OUT) and isinstance(exception_type, CancelledError):
            raise TimeoutError from None
        
        return False
    
    
    def __repr__(self):
        """Returns the timeout's representation."""
        return f'<{type(self).__name__}>'


def set_tcp_nodelay(transport, value):
    """
    Sets or removes tcp nodelay socket option to the given transport's socket if applicable.
    
    Parameters
    ----------
    transport : ``AbstractTransportLayerBase``
        Asynchronous transport implementation.
    value : `bool`
        Value to set tcp nodelay to.
    """
    socket = transport.get_extra_info('socket')
    if socket is None:
        return
    
    if socket.family not in (module_socket.AF_INET, module_socket.AF_INET6):
        return
    
    # socket may be closed already, on windows OSError get raised
    try:
        socket.setsockopt(module_socket.IPPROTO_TCP, module_socket.TCP_NODELAY, value)
    except OSError:
        pass


def freeze_headers(headers):
    """
    Creates a frozen version of the given headers.
    
    Parameters
    ----------
    headers : `None | IgnoreCaseMultiValueDictionary`
        Proxy headers.
    
    Returns
    -------
    proxy_headers_frozen : `None | tuple<(str, tuple<str>)>`
    """
    if (headers is not None) and headers:
        return (*((key, tuple(headers.get_all(key, ()))) for key in sorted(headers.keys())),)


HTTP_STATUS_RP = re_compile(b'[ \t]*HTTP/(\\d+)\\.(\\d+)[ \t]+(\\d\\d\\d)(?:[ \t]+(.*?))?[ \t]*(?:\r\n|$)')
HTTP_REQUEST_RP = re_compile(b'[ \t]*([^ \t]+)[ \t]+([^ \t]+)[ \t]+HTTP/(\\d+)\\.(\\d+)[ \t]*(?:\r\n|$)')


def parse_http_response(data):
    """
    Parses http response message.
    
    Parameters
    ----------
    data : `bytes`
        Data to parse from.
    
    Returns
    -------
    response_message : ``RawResponseMessage``
    
    Raises
    ------
    PayloadError
        Invalid data received.
    """
    parsed = HTTP_STATUS_RP.match(data, 0)
    if parsed is None:
        line_end = data.find(b'\r\n')
        if line_end == -1:
            line_end = len(data)
        
        raise PayloadError(f'Invalid status line: {data[0 : line_end]!r}.')
    
    offset = parsed.end()
    major, minor, status, reason = parsed.groups()
    if reason is None:
        pass
    if not reason:
        reason = None
    else:
        reason = reason.decode('utf-8', errors = 'surrogateescape')
    
    headers = parse_http_headers(data, offset)
    return RawResponseMessage(HttpVersion(int(major), int(minor)), int(status), reason, headers)


def parse_http_request(data):
    """
    Parses http request message.
    
    Parameters
    ----------
    data : `bytes`
        Data to parse from.
    
    Returns
    -------
    request_message : ``RawRequestMessage``
    
    Raises
    ------
    PayloadError
        Invalid data received.
    """
    parsed = HTTP_REQUEST_RP.match(data, 0)
    if parsed is None:
        line_end = data.find(b'\r\n')
        if line_end == -1:
            line_end = len(data)
        
        raise PayloadError(f'Invalid request line: {data[0 : line_end]!r}.')
    
    offset = parsed.end()
    method, path, major, minor = parsed.groups()
    method = method.decode('utf-8', 'surrogateescape').upper()
    path = path.decode('utf-8', 'surrogateescape')

    headers = parse_http_headers(data, offset)
    return RawRequestMessage(HttpVersion(int(major), int(minor)), method, path, headers)


def parse_http_headers(data, offset):
    """
    Parses http headers from the given data.
    
    Parameters
    ----------
    data : `bytes`
        Data to parse from.
    
    offset : `int`
        Position to start parsing at.
    
    Returns
    -------
    headers : ``IgnoreCaseMultiValueDictionary``
    
    Raises
    ------
    PayloadError
        Invalid data received.
    """
    headers = IgnoreCaseMultiValueDictionary()
    while True:
        end_index = data.find(b'\r\n', offset)
        if end_index == -1:
            end_index = len(data)
        
        # Found \r\n instantly, we done!
        # The case, when there is nothing inside of the headers.
        if end_index <= offset:
            return headers
        
        # At this case something is in the line, but is it a header?
        middle_index = data.find(b':', offset, end_index)
        if middle_index <= offset:
            raise PayloadError(f'Invalid header line: {data[offset : end_index]!r}.')
        
        name = data[offset : middle_index].lstrip()
        value = data[middle_index + 1 : end_index].strip()
        offset = end_index + 2
            
        name = name.decode('utf-8', 'surrogateescape')
        
        # continuous?
        if (len(data) > offset) and data[offset] in (b'\t'[0], b' '[0]):
            value_continuous = [value]
            while True:
                end_index = data.find(b'\r\n', offset)
                if end_index == -1:
                    end_index = len(data)
                
                # find \r\n
                value_continuous.append(data[offset : end_index].strip())
                # add \r\n shift
                offset = end_index + 2
                
                # continuous again?
                if (len(data) > offset) and data[offset] in (b'\t'[0], b' '[0]):
                    continue
                break
            
            value = b' '.join(value_continuous)
            value_continuous = None
        
        # Store, nice!
        headers[name] = value.decode('utf-8', 'surrogateescape')



__all__ = ()

import sys, re, termios, tty
from os import get_blocking, set_blocking, write
from select import poll as Poller, POLLOUT as EVENT_POLL_WRITE
from selectors import DefaultSelector, EVENT_READ
from socket import socketpair as create_socket_pair
from time import monotonic

from ....utils import copy_docs

from ...keys import (
    KEY_ARROW_DOWN, KEY_ARROW_END, KEY_ARROW_HOME, KEY_ARROW_LEFT, KEY_ARROW_RIGHT, KEY_ARROW_UP, KEY_BACK_TAB,
    KEY_DELETE_LEFT, KEY_DELETE_RIGHT, KEY_DESTROY, KEY_KEYBOARD_INTERRUPT, KEY_NEW_LINE_ALL, KEY_TAB
)

from .compilation import maybe_compile
from .display_state import DisplayState, get_new_content_width
from .editor_base import EditorBase
from .prefix_trimming import trim_console_prefix
from .terminal_control_commands import (
    COMMAND_FORMAT_RESET, COMMAND_DOWN, COMMAND_UP, COMMAND_START_LINE
)


DEDENT_WORDS = frozenset((
    'pass',
    'return',
    'continue',
    'break'
))

EMPTY_CHARACTERS = frozenset((' ', '\t', '\n'))
AUTO_COMPLETE_BREAK_CHARACTERS = frozenset((
    ' ', '"', '#', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '@', '[', '\\', ']',
    '{', '|', '}', '~'
))


INDEXED_INPUT_RP = re.compile('\\s*in\\s*\\[\\s*(\\d+)\\s*\\]\\s*', re.I)


WRITE_TIMEOUT = 30.0


class KeyNode:
    """
    A key's node when parsing input stream.
    
    Attributes
    ----------
    character_sub_nodes : `None | dict<str, instance>`
        Sub-nodes, character to sub-node relation.
    end : `bool`
        Whether the node is an acceptable end.
    matcher_sub_nodes : `None | list<(FunctionType, instance)>`
        Matchers, function to sub-node relation.
    """
    __slots__ = ('character_sub_nodes', 'end', 'matcher_sub_nodes')
    
    def __new__(cls):
        """
        Creates a new key node instance.
        """
        self = object.__new__(cls)
        self.character_sub_nodes = None
        self.end = False
        self.matcher_sub_nodes = None
        return self
    
    
    def __repr__(self):
        """Returns the key node's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # end
        end = self.end
        if end:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' end = ')
            repr_parts.append(repr(end))
        
        # character_sub_nodes
        character_sub_nodes = self.character_sub_nodes
        if (character_sub_nodes is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' character_sub_nodes = ')
            repr_parts.append(repr(character_sub_nodes))
        
        # matcher_sub_nodes
        matcher_sub_nodes = self.matcher_sub_nodes
        if (matcher_sub_nodes is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' matcher_sub_nodes = ')
            repr_parts.append(repr(matcher_sub_nodes))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two key nodes are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.character_sub_nodes != other.character_sub_nodes:
            return False
        
        if self.end != other.end:
            return False
        
        if self.matcher_sub_nodes != other.matcher_sub_nodes:
            return False
        
        return True
    
    
    def set_end(self):
        """
        Sets ``.end`` to true.
        """
        self.end = True
    
    
    def add_character_sub_node(self, character):
        """
        Adds a character sub-node.
        
        Parameters
        ----------
        character : `str`
            Character to add.
        
        Returns
        -------
        node : `instance<type<self>>`
        """
        character_sub_nodes = self.character_sub_nodes
        if character_sub_nodes is None:
            character_sub_nodes = {}
            self.character_sub_nodes = character_sub_nodes
            node = None
        
        else:
            node = character_sub_nodes.get(character, None)
        
        if node is None:
            node = type(self)()
            character_sub_nodes[character] = node
        
        return node
    
    
    def add_matcher_sub_node(self, matcher):
        """
        Adds a matcher sub-node.
        
        Parameters
        ----------
        matcher : `FunctionType`
            Matcher to add.
        
        Returns
        -------
        node : `instance<type<self>>`
        """
        matcher_sub_nodes = self.matcher_sub_nodes
        if matcher_sub_nodes is None:
            matcher_sub_nodes = []
            self.matcher_sub_nodes = matcher_sub_nodes
            node = None
        else:
            for iterated_matched, iterated_node in matcher_sub_nodes:
                if iterated_matched is matcher:
                    node = iterated_node
                    break
            else:
                node = None
        
        if node is None:
            node = type(self)()
            matcher_sub_nodes.append((matcher, node))
        
        return node
    
    
    def match(self, character):
        """
        Tries to match the given character.
        
        Parameters
        ----------
        character : `bool`
            The character to match.
        
        Returns
        -------
        node : `Node | KeyNode`
        """
        character_sub_nodes = self.character_sub_nodes
        if (character_sub_nodes is not None):
            try:
                node = character_sub_nodes[character]
            except KeyError:
                pass
            else:
                return node
            
        matcher_sub_nodes = self.matcher_sub_nodes
        if (matcher_sub_nodes is not None):
            for matcher, node in matcher_sub_nodes:
                if matcher(character):
                    return node 
        
        return None


def _add_keys(core_node, keys):
    """
    Adds a key to a node map.
    
    Parameters
    ----------
    core_node : ``KeyNode``
        Key node to add the keys to.
    keys : `list<str>`
        Keys to build the node map with.
    """
    for key in keys:
        node = core_node
        for character in key:
            node = node.add_character_sub_node(character)
        
        node.set_end()


NODE_MAP = KeyNode()
_add_keys(
    NODE_MAP,
    [
        KEY_KEYBOARD_INTERRUPT,
        KEY_DESTROY,
        KEY_TAB,
        *KEY_NEW_LINE_ALL,
        KEY_ARROW_UP,
        KEY_ARROW_DOWN,
        KEY_ARROW_RIGHT,
        KEY_ARROW_LEFT,
        KEY_ARROW_END,
        KEY_ARROW_HOME,
        KEY_BACK_TAB,
        KEY_DELETE_RIGHT,
        KEY_DELETE_LEFT,
    ],
)
NODE_MAP.add_matcher_sub_node(str.isprintable).set_end()


def _read_from_node_map(node, input_stream):
    """
    Reads the next token from a node map.
    
    Parameters
    ----------
    node : `KeyNode`
        The node to read from.
    input_stream : ``InputStream``
        Stream to read from.
    
    Returns
    -------
    output : `None | str`
    """
    end = 0
    peek_count = 0
    
    while True:
        character = input_stream.peek(peek_count)
        if character is None:
            break
        
        matched_node = node.match(character)
        if matched_node is None:
            break
        
        peek_count += 1
        node = matched_node
        
        if node.end:
            end = peek_count
        
        continue
    
    if end <= 0:
        return None
    
    output = input_stream.read(end)
    input_stream.consume(end)
    return output



def get_starting_space_count(line):
    """
    Returns how much space the line has at it's start.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    
    Returns
    -------
    space_count : `int`
    """
    space_count = 0
    for character in line:
        if character != ' ':
            break
            
        space_count += 1
        continue
    
    return space_count



def line_ends_with(line, target):
    """
    Returns whether the given line ends with the given character.
    
    Spaces ignored.
    
    Parameters
    ----------
    line : `str`
        The lien to check out.
    target : `str`
        The character to look for.
    
    Returns
    -------
    ends_with : `bool`
    """
    for character in reversed(line):
        if character == target:
            return True
        
        if character == ' ':
            continue
        
        break
    
    return False


def line_starts_with_word_any(line, words):
    """
    Returns whether the given line ends with the any of the given words.
    
    Spaces ignored.
    
    Parameters
    ----------
    line : `str`
        The lien to check out.
    words : `frozenset` of `str`
        The words to look for.
    
    Returns
    -------
    ends_with : `bool`
    """
    line = line.lstrip()
    
    for word in words:
        if line.startswith(word):
            if len(line) == len(word) or line[len(word)] == ' ':
                return True
    
    return False


def _check_is_line_empty(line):
    """
    Returns whether the given line is empty.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    
    Returns
    -------
    is_empty : `bool`
    """
    for character in line:
        if character not in EMPTY_CHARACTERS:
            return False
    
    return True


def _check_is_line_empty_at(line, index):
    """
    Returns whether the line is empty at the given position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The exact position.
    
    Returns
    -------
    is_line_empty_at : `bool`
    """
    line_length = len(line)
    if (index < line_length) and (index >= 0):
        if line[index] not in EMPTY_CHARACTERS:
            return False
    
    return True


def _check_is_buffer_line_empty_at(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty at the given index.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The exact position.
    
    Returns
    -------
    is_buffer_line_empty_at : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_at(buffer[line_index], index)
    
    return True


def _check_is_line_empty_after(line, index):
    """
    Returns whether the line is empty after the given index. Excluding the character at `index` position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_line_empty_after : `bool`
    """
    index += 1
    
    line_length = len(line)
    while index < line_length:
        if line[index] not in EMPTY_CHARACTERS:
            return False
        
        index += 1
    
    return True


def _check_is_buffer_line_empty_after(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty after the given index.
    Excluding the character at `index` position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_line_empty_after : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_after(buffer[line_index], index)
    
    return True


def _check_is_line_empty_before(line, index):
    """
    Returns whether the line is empty before the given index. Excluding the character at `index` position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_line_empty_before : `bool`
    """
    index -= 1
    
    line_length = len(line)
    if index >= line_length:
        index = line_length - 1
    
    while index >= 0:
        if line[index] not in EMPTY_CHARACTERS:
            return False
        
        index -= 1
        continue

    return True


def _check_is_buffer_line_empty_before(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty before the given index.
    Excluding the character at `index` position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_line_empty_before : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_before(buffer[line_index], index)
    
    return True


def _check_is_buffer_empty_at(buffer, index):
    """
    Returns whether the exact line of the buffer is empty.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The line's index to check out.
    
    Returns
    -------
    is buffer_empty_at : `bool`
    """
    buffer_length = len(buffer)
    if (index < buffer_length) and (index >= 0):
        return _check_is_line_empty(buffer[index])
    
    return True


def _check_is_buffer_empty_before(buffer, index):
    """
    Returns whether the buffer is empty before the given index. Excluding the `index`'s exact position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_empty_before : `bool`
    """
    index -= 1
    
    buffer_length = len(buffer)
    if index >= buffer_length:
        index = buffer_length - 1
    
    
    while index >= 0:
        if not _check_is_line_empty(buffer[index]):
            return False
        
        index -= 1
        continue
    
    return True


def _check_is_buffer_empty_after(buffer, index):
    """
    Returns whether the buffer is empty after the given index. Excluding the `index`'s exact position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_empty_after : `bool`
    """
    buffer_length = len(buffer)
    
    index += 1
    
    while index < buffer_length:
        if not _check_is_line_empty(buffer[index]):
            return False
        
        index += 1
        continue
    
    return True


def _check_should_try_compile_one_line(display_state):
    """
    Checks whether the current input should be tried to be compiled.
    
    This function is called when the buffer's length is `exactly 1`.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    line = display_state.buffer[display_state.cursor_line_index]
    cursor_index = display_state.cursor_index
    
    if not _check_is_line_empty_after(line, cursor_index):
        return False
    
    if not _check_is_line_empty_at(line, cursor_index):
        return False
    
    if _check_is_line_empty_before(line, cursor_index):
        return False
    
    return True


def _check_should_try_compile_multi_line(display_state):
    """
    Checks whether the current input should be tried to be compiled.
    
    This function is called when the buffer's length is `over 1`.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    buffer = display_state.buffer
    
    # If we are checking multi-line
    
    if len(buffer) < 3:
        return False
    
    cursor_index = display_state.cursor_index
    cursor_line_index = display_state.cursor_line_index
    
    # If there is anything after our current position return false.
    
    if (
        not _check_is_buffer_line_empty_at(buffer, cursor_line_index, cursor_index) or
        not _check_is_buffer_line_empty_after(buffer, cursor_line_index, cursor_index) or
        not _check_is_buffer_empty_after(buffer, cursor_line_index)
    ):
        return False
    
    # If there is no code before the current position return false.
    
    if (
        _check_is_buffer_line_empty_before(buffer, cursor_line_index, cursor_index) and
        _check_is_buffer_empty_before(buffer, cursor_line_index)
    ):
        return False
    
    # We should only compile when we have 1 free line above us and our current line is empty too.
    if (
        _check_is_buffer_empty_after(buffer, cursor_line_index) and
        _check_is_buffer_empty_at(buffer, cursor_line_index) and
        _check_is_buffer_empty_at(buffer, cursor_line_index - 1)
    ):
        return True
    
    return False


def _check_should_try_compile(display_state):
    """
    Tries whether based on the current input, we should try to compile the input.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    buffer_length = len(display_state.buffer)
    
    if buffer_length == 0:
        # Should not happen
        return False
    
    if buffer_length == 1:
        return _check_should_try_compile_one_line(display_state)
    
    return _check_should_try_compile_multi_line(display_state)


def _get_identifier_to_autocomplete(line, index):
    """
    Gets identifier to autocomplete.
    
    Parameters
    ----------
    line : `str`
        Current line.
    index : `int`
        The cursor's index in the line.
    
    Returns
    -------
    identifier_to_autocomplete : `None`, `str`
    """
    line_length = len(line)
    # We only want to autocomplete if the end does not contain identifier characters.
    if line_length > index:
        if line[index] not in AUTO_COMPLETE_BREAK_CHARACTERS:
            return None
    
    # Collect the last word
    last_word_parts = []
    
    for index__in_loop in reversed(range(0, index)):
        character = line[index__in_loop]
        if character in AUTO_COMPLETE_BREAK_CHARACTERS:
            break
        
        last_word_parts.append(character)
        continue
    
    # Nothing to auto complete
    if not last_word_parts:
        return None
    
    # Do not auto complete if we are in an attribute access operation
    for index__in_loop in reversed(range(0, index - len(last_word_parts))):
        character = line[index__in_loop]
        if character == '.':
            return None
        
        if character == ' ':
            continue
        
        break
    
    # Last word should be an identifier
    last_word_parts.reverse()
    last_word = ''.join(last_word_parts)
    if not last_word.isidentifier():
        return None
    
    return last_word


class InputStream:
    """
    Helper class for processing a bigger chunk of input data at once.
    
    Attributes
    ----------
    content : `str`
        Content to iterate over.
    index : `str`
        The last read index.
    """
    __slots__ = ('content', 'index')
    
    def __new__(cls, content):
        """
        Creates a new input iterator.
        
        Parameters
        ----------
        content : `str`
            The content to iterate over.
        """
        self = object.__new__(cls)
        self.content = content
        self.index = 0
        return self
    
    
    def __repr__(self):
        """Returns the input stream's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' index = ')
        repr_parts.append(repr(self.index))
        
        repr_parts.append(', content = ')
        repr_parts.append(repr(self.content))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two input streams are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.content != other.content:
            return False
        
        if self.index != other.index:
            return False
        
        return True
    
    
    def peek(self, count):
        """
        Peeks forward or backwards. Defaults to not peeking, but getting the current character.
        
        Parameters
        ----------
        count : `int`
            How much to peek.
        
        Returns
        -------
        character : `None | str`
        """
        content = self.content
        index = self.index + count
        if (index < 0) or (index >= len(content)):
            return None
        
        return content[index]
    
    
    def consume(self, count):
        """
        Consumes characters.
        
        Parameters
        ----------
        count : `int`
            How much to consume.
        """
        if count < 1:
            return
        
        # limit index to length, so it cant go wild.
        index = self.index + count
        length = len(self.content)
        if index > length:
            index = length
        self.index = index
    
    
    def read(self, count):
        """
        Reads from the stream from the current position.
        
        Parameters
        ----------
        count : `int
            How much to read.
        
        Returns
        -------
        output : `None | str`
        """
        if count <= 0:
            return None
        
        content = self.content
        index = self.index
        if index >= len(content):
            return None
        
        return content[index : index + count]
    
    
    def is_at_end_of_stream(self):
        """
        Returns whether the editor at the end of the stream.
        
        Returns
        -------
        is_at_end_of_stream : `bool` 
        """
        return self.index >= len(self.content)
    
    
    def copy(self):
        """
        Copies the input stream.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.content = self.content
        new.index = self.index
        return new


class InputTokenizer:
    """
    Creates standalone tokens from an input stream. Nothing fancy.
    
    Attributes
    ----------
    stream : ``InputStream``
        Stream to use.
    """
    __slots__ = ('stream',)
    
    def __new__(cls, stream):
        """
        Creates a new input tokenizer.
        
        Parameters
        ----------
        stream : ``InputStream``
            Stream to use.
        """
        self = object.__new__(cls)
        self.stream = stream
        return self
    
    
    def __repr__(self):
        """Returns the input tokenizer's representation."""
        return ''.join(['<', self.__class__.__name__, ' stream = ', repr(self.stream), '>'])
    
    
    def next(self):
        """
        Returns the next token in the tokenizer.
        
        Returns
        -------
        token : `None | str`
            Returns `None` if there are no more tokens to read.
        """
        stream = self.stream
        while not stream.is_at_end_of_stream():
            output = _read_from_node_map(NODE_MAP, stream)
            if output is not None:
                return output
            
            stream.consume(1)
            continue
    
    
    def is_length_greater_than(self, value):
        """
        Returns whether the tokenizer's length is greater than the given value.
        
        Parameters
        ----------
        value : `bool`
            The value to check for.
        
        Returns
        -------
        is_length_greater_than : `bool`
        """
        stream = self.stream
        index = stream.index
        
        while value >= 0:
            token = self.next()
            if token is None:
                outcome = False
                break
            
            value -= 1
            continue
        
        else:
            outcome = True
            
        stream.index = index
        return outcome


def try_decode(data):
    """
    Tries to decode the given bytes data. On decode error changes decode error handling for that specific part.
    
    Parameters
    ----------
    data : `bytes`
        The data to decode.
    
    Returns
    -------
    content : `str`
        The decoded content.
    data_continuous : `None`, `bytes`
        Undecodable data at the end.
    """
    content_chunk_parts = None
    data_continuous = None
    content = ''
    
    while True:
        try:
            content = data.decode()
        except UnicodeDecodeError as err:
            
            end = err.end
            start = err.start
            length = len(data)
            
            if end == length:
                if start > 0:
                    content = data[:start].decode()
                    
                    if (content_chunk_parts is not None):
                        content_chunk_parts.append(content)
                
                data_continuous = data[start:]
                break
            
            if content_chunk_parts is None:
                content_chunk_parts = []
            
            if start > 0:
                content_chunk_parts.append(data[:start].decode())
            
            content_chunk_parts.append(data[start:end].decode(errors = 'backslashreplace'))
            data = data[end:]
            continue
        
        else:
            if (content_chunk_parts is not None):
                content_chunk_parts.append(content)
            break
    
    
    if content_chunk_parts is None:
        content = content
    
    else:
        content = ''.join(content_chunk_parts)
    
    return content, data_continuous


class EditorAdvanced(EditorBase):
    """
    A simple repl command editor.
    
    Attributes
    ----------
    alive : `str`
        Whether the input loop is alive.
    async_output_content_continuous : `None`, `str`
        Async content to write. Set if the last async content chunk did not end with linebreak.
    async_output_data_continuous : `None`, `bytes`
        Async content to write. Set if the end of content is not decodable.
    async_output_written : `bool`
        Whether async output was already written.
    auto_completer : ``AutoCompleter``
        Auto completer to use.
    compiled_code : `None`, ``CodeType``
        The compiled inputted code.
    display_state : ``DisplayState``
        The currently displayed state.
    file_name : `str`
        File name of the code produced by the editor.
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    history : ``History``
        History used for caching inputs.
    input_stream : `file-like`
        Input stream to read.
    input_stream_blocking_original : `bool`
        The original value of input stream's blockedness.
    input_stream_settings_original : `list` (`int`, `int`, `int`, `int`, `int`, `int`, `list` of `bytes`)
        The original settings og the input stream.
    modified : `bool`
        Whether the input is modified.
    output_proxy_read_socket : `Socket`
        Read pair socket of ``.output_proxy_write_socket``.
    output_proxy_write_socket : `Socket`
        Socket proxying stdout and stderr.
    output_stream : `file-like`
        Output stream to write.
    prefix_continuous : `str`
        Non-first line's prefix.
    prefix_initial : `str`
        First line's prefix.
    prefix_length : `int`
        As how long should the prefix's length be interpreted
    selector : ``DefaultSelector``
        Selector used to poll from `stdin`, `stdout` and `stderr`.
    """
    __slots__ = (
        'alive', 'async_output_content_continuous', 'async_output_data_continuous', 'async_output_written',
        'display_state', 'file_name', 'highlighter', 'history', 'input_stream', 'input_stream_blocking_original',
        'input_stream_settings_original', 'modified', 'output_proxy_read_socket', 'output_proxy_write_socket',
        'output_stream', 'selector'
    )
    
    @copy_docs(EditorBase.__new__)
    def __new__(
        cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history, auto_completer
    ):
        
        display_state = DisplayState(buffer)
        
        input_stream = sys.stdin
        input_stream_file_descriptor = input_stream.fileno()
        input_stream_blocking_original = get_blocking(input_stream_file_descriptor)
        set_blocking(input_stream_file_descriptor, False)
        input_stream_settings_original = termios.tcgetattr(input_stream)
        
        output_proxy_read_socket, output_proxy_write_socket = create_socket_pair()
        set_blocking(output_proxy_read_socket.fileno(), False)
        
        
        self = EditorBase.__new__(
            cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history,
            auto_completer
        )
        
        self.alive = False
        self.async_output_content_continuous = None
        self.async_output_data_continuous = None
        self.async_output_written = False
        self.display_state = display_state
        self.history = history
        self.input_stream = input_stream
        self.input_stream_blocking_original = input_stream_blocking_original
        self.input_stream_settings_original = input_stream_settings_original
        self.modified = False
        self.output_proxy_read_socket = output_proxy_read_socket
        self.output_proxy_write_socket = output_proxy_write_socket
        self.output_stream = sys.stdout
        self.selector = DefaultSelector()
        
        return self
    
    
    @copy_docs(EditorBase.__call__)
    def __call__(self):
        new_stdout = open(self.output_proxy_write_socket.fileno(), 'w')
        old_stderr = sys.stderr
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        
        tty.setraw(self.input_stream)
        self.selector.register(self.input_stream.fileno(), EVENT_READ)
        self.selector.register(self.output_proxy_read_socket.fileno(), EVENT_READ)
        
        self.output_stream.flush()
        
        try:
            self.alive = True
            self.initialise_display()
            
            while True:
                if not self.alive:
                    break
                
                new_display_state = self.poll_and_process_input()
                if (new_display_state is not None):
                    self.update_display(new_display_state)
        
        finally:
            self.alive = False
            self._maybe_display_continuous_content()
            
            write_buffer = []
            self.display_state.jump_to_end(
                write_buffer, self.prefix_length, get_new_content_width(self.prefix_length)
            )
            write_buffer.append(COMMAND_DOWN)
            write_buffer.append(COMMAND_START_LINE)
            write_buffer.append(COMMAND_FORMAT_RESET)
            try:
                self._write_write_buffer(write_buffer)
            finally:
                write_buffer = None
            
            termios.tcsetattr(self.input_stream, termios.TCSADRAIN, self.input_stream_settings_original)
            set_blocking(self.input_stream.fileno(), self.input_stream_blocking_original)
            
            sys.stdout = self.output_stream
            sys.stderr = old_stderr
            
            try:
                self.output_proxy_read_socket.close()
            except OSError as err:
                if err.errno != 9:
                    raise
            finally:
                self.output_proxy_read_socket = None
            
            try:
                self.output_proxy_write_socket.close()
            except OSError as err:
                if err.errno != 9:
                    raise
            finally:
                self.output_proxy_write_socket = None
            
            self.selector.close()
        
        return self.compiled_code
    
    
    @copy_docs(EditorBase.get_buffer)
    def get_buffer(self):
        return self.display_state.buffer
    
    
    @classmethod
    @copy_docs(EditorBase.has_history_support)
    def has_history_support(cls):
        return True
    
    
    def execute_enter(self, old_display_state, should_auto_format):
        """
        Executes an enter key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether auto formatting should be applied.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        
        Raises
        ------
        SyntaxError
        """
        if should_auto_format:
            new_display_state = self.checkout_indexed_input(old_display_state)
            if (new_display_state is not None):
                return new_display_state
            
            if self.check_exit_conditions(old_display_state):
                return None
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        current_line_new = line[:cursor_index]
        next_line = line[cursor_index:]
        
        if should_auto_format:
            new_line_starter_space_count = get_starting_space_count(line)
            # In which order should we check this?
            if line_ends_with(current_line_new, ':'):
                new_line_starter_space_count += 4
            elif line_starts_with_word_any(current_line_new, DEDENT_WORDS):
                new_line_starter_space_count -= 4
            
            if new_line_starter_space_count < 0:
                new_line_starter_space_count = 0
            
            next_line = ' ' * new_line_starter_space_count + next_line
        
        else:
            new_line_starter_space_count = 0
        
        buffer[cursor_line_index] = current_line_new
        cursor_line_index += 1
        buffer.insert(cursor_line_index, next_line)
        
        new_display_state.cursor_index = new_line_starter_space_count
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_arrow_left(self, old_display_state):
        """
        Executes an arrow-left key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if (cursor_index == 0):
            if (cursor_line_index == 0):
                return None
            
            cursor_line_index -= 1
            cursor_index = len(buffer[cursor_line_index])
        
        else:
            cursor_index -= 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        return new_display_state
    
    
    def execute_arrow_home(self, old_display_state):
        """
        Executes a `home` key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        cursor_index = new_display_state.cursor_index
        
        if (cursor_index == 0):
            return None
        
        cursor_index = 0
        
        new_display_state.cursor_index = cursor_index
        return new_display_state
        
    
    def execute_arrow_right(self, old_display_state):
        """
        Executes an arrow-right key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line_length = len(buffer[cursor_line_index])
        if cursor_index == line_length:
            if cursor_line_index == len(buffer) - 1:
                return None
            
            cursor_line_index += 1
            cursor_index = 0
        else:
            cursor_index += 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        return new_display_state
    
    
    def execute_arrow_end(self, old_display_state):
        """
        Executes an `end` key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line_length = len(buffer[cursor_line_index])
        if cursor_index == line_length:
            return None
        
        cursor_index = line_length
        
        new_display_state.cursor_index = cursor_index
        return new_display_state
    
    
    def execute_arrow_up(self, old_display_state, should_auto_format):
        """
        Executes an arrow-up key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if should_auto_format:
            new_display_state = self.try_rewind_in_history()
            if (new_display_state is not None):
                return new_display_state
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_line_index == 0:
            if (cursor_index == 0):
                return None
            
            cursor_index = 0
        
        else:
            cursor_line_index -= 1
            
            line_length = len(buffer[cursor_line_index])
            if line_length < cursor_index:
                cursor_index = line_length
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        return new_display_state
    
    
    def execute_arrow_down(self, old_display_state, should_auto_format):
        """
        Executes an arrow-down key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if should_auto_format:
            new_display_state = self.try_forward_in_history()
            if (new_display_state is not None):
                return new_display_state
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_line_index == len(buffer) - 1:
            line_length = len(buffer[cursor_line_index])
            if cursor_index == line_length:
                return None
            
            cursor_index = line_length
        
        else:
            cursor_line_index += 1
            
            line_length = len(buffer[cursor_line_index])
            if line_length < cursor_index:
                cursor_index = line_length
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        return new_display_state
    
    
    def execute_arrow_back_tab(self, old_display_state):
        """
        Executes a back-tab key press (shift + tab).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        
        back_space_count = get_starting_space_count(line)
        if back_space_count == 0:
            return None
        
        remove_space_count = 4 - (-back_space_count & 3)
        cursor_index -= remove_space_count
        if (cursor_index < 0):
            cursor_index = 0
        line = line[remove_space_count:]
        buffer[cursor_line_index] = line
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_tab(self, old_display_state, should_auto_format):
        """
        Executes a tab key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether auto formatting should be applied.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        line = buffer[cursor_line_index]
        
        # Use goto
        while True:
            if should_auto_format:
                identifier_to_auto_complete = _get_identifier_to_autocomplete(line, cursor_index)
                if (identifier_to_auto_complete is not None):
                    common_prefix = self.auto_completer.get_common_prefix(identifier_to_auto_complete)
                    if (common_prefix is None):
                        modified = False
                        break
                    
                    added_content = common_prefix[len(identifier_to_auto_complete):]
                    buffer[cursor_line_index] = f'{line[:cursor_index]}{added_content}{line[cursor_index:]}'
                    cursor_index += len(added_content)
                    modified = True
                    break
            
            if should_auto_format:
                add_spaces_count =  4 - (cursor_index & 3)
                line_length = len(line)
                while add_spaces_count:
                    if cursor_index >= line_length:
                        break
                    
                    if line[cursor_index] != ' ':
                        break
                    
                    cursor_index += 1
                    add_spaces_count -= 1
            else:
                add_spaces_count = 4
            
            if add_spaces_count:
                buffer[cursor_line_index] = f'{line[:cursor_index]}{" " * add_spaces_count}{line[cursor_index:]}'
                cursor_index += add_spaces_count
                modified = True
                break
            
            modified = False
            break
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = modified
        
        return new_display_state
    
    
    def execute_delete_left(self, old_display_state):
        """
        Executes a delete-left key press (backspace).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_index == 0:
            if cursor_line_index == 0:
                return None
                
            line = buffer.pop(cursor_line_index)
            cursor_line_index -= 1
            previous_line = buffer[cursor_line_index]
            buffer[cursor_line_index] = previous_line + line
            cursor_index = len(previous_line)
        
        else:
            line = buffer[cursor_line_index]
            buffer[cursor_line_index] = f'{line[:cursor_index - 1]}{line[cursor_index:]}'
            cursor_index -= 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_delete_right(self, old_display_state):
        """
        Executes a delete-right key press (delete).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        line_length = len(line)
        if cursor_index == line_length:
            if cursor_line_index == len(buffer) - 1:
                return None
                
            next_line = buffer.pop(cursor_line_index + 1)
            line = line + next_line
            buffer[cursor_line_index] = line
                
        else:
            line = f'{line[:cursor_index]}{line[cursor_index + 1:]}'
            buffer[cursor_line_index] = line
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_print(self, old_display_state, character_string):
        """
        Executes a printable character key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        character_string : `str`
            The character to print out.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        buffer[cursor_line_index] = f'{line[:cursor_index]}{character_string}{line[cursor_index:]}'
        cursor_index += 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def poll(self):
        """
        Polls a character from any input stream.
        
        Returns
        -------
        content : `str`
            The inputted string into `sys.stdin`.
        """
        content = None
        
        while True:
            # We force flush every second.
            sys.stdout.flush()
            
            event_list = self.selector.select(1)
            if not event_list:
                continue
                
            for key, mask in event_list:
                file_descriptor = key.fileobj
                
                if (file_descriptor == self.output_proxy_read_socket.fileno()) and (mask & EVENT_READ):
                    try:
                        data = self.output_proxy_read_socket.recv(1024)
                    except (BlockingIOError, InterruptedError):
                        pass
                    else:
                        if data:
                            self._process_async_content(data)
                        
                        data = None
                    
                    continue
                
                if (file_descriptor == self.input_stream.fileno()) and (mask & EVENT_READ):
                    try:
                        content = self.input_stream.read()
                    except (BlockingIOError, InterruptedError):
                        pass
                    
                    continue
            
            
            if (content is not None):
                if content:
                    return content
                
                content = None
    
    
    def poll_and_process_input(self):
        """
        Inputs one action and returns whether any actions took place.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        
        Raises
        ------
        KeyboardInterrupt
        SystemExit
        SyntaxError
        """
        content = self.poll()
        
        input_stream = InputStream(content)
        input_tokenizer = InputTokenizer(input_stream)
        if not input_tokenizer.is_length_greater_than(1):
            should_auto_format = True
        else:
            trimmed_content = trim_console_prefix(content)
            if (trimmed_content is None):
                should_auto_format = False
            
            else:
                input_stream = InputStream(trimmed_content)
                input_tokenizer = InputTokenizer(input_stream)
                should_auto_format = not input_tokenizer.is_length_greater_than(1)
        
        old_display_state = self.display_state
        new_display_state = old_display_state
        
        while True:
            token = input_tokenizer.next()
            if token is None:
                break
            
            received_display_state = self.process_input(new_display_state, token, should_auto_format)
            if (received_display_state is not None):
                new_display_state = received_display_state
            continue
        
        if (new_display_state is old_display_state):
            new_display_state = None
        
        return new_display_state
    
    
    def process_input(self, old_display_state, token, should_auto_format):
        """
        Processes one input.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        token : `str`
            A single input string.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if token == KEY_KEYBOARD_INTERRUPT:
            raise KeyboardInterrupt()
        
        if token == KEY_DESTROY:
            raise SystemExit()
        
        if token in KEY_NEW_LINE_ALL:
            return self.execute_enter(old_display_state, should_auto_format)
    
        if token == KEY_ARROW_LEFT:
            return self.execute_arrow_left(old_display_state)
        
        if token == KEY_ARROW_RIGHT:
            return self.execute_arrow_right(old_display_state)
        
        if token == KEY_ARROW_UP:
            return self.execute_arrow_up(old_display_state, should_auto_format)
        
        if token == KEY_ARROW_DOWN:
            return self.execute_arrow_down(old_display_state, should_auto_format)
        
        if token == KEY_ARROW_END:
            return self.execute_arrow_end(old_display_state)
        
        if token == KEY_ARROW_HOME:
            return self.execute_arrow_home(old_display_state)
        
        if token == KEY_BACK_TAB:
            return self.execute_arrow_back_tab(old_display_state)
        
        if token == KEY_DELETE_RIGHT:
            return self.execute_delete_right(old_display_state)
        
        if token == KEY_TAB:
            return self.execute_tab(old_display_state, should_auto_format)
        
        if token == KEY_DELETE_LEFT:
            return self.execute_delete_left(old_display_state)
        
        if token.isprintable():
            return self.execute_print(old_display_state, token)
        
        return None
    
    
    def initialise_display(self):
        """
        Initialises the display.
        
        Called after the editor was started.
        """
        content_width = get_new_content_width(self.prefix_length)
        write_buffer = []
        self.display_state.write(
            write_buffer, self.prefix_length, self.prefix_initial, self.prefix_continuous, content_width
        )
        try:
            self._write_write_buffer(write_buffer)
        finally:
            write_buffer = None
    
    
    def update_display(self, new_display_state):
        """
        Updates the display.
        
        This method is called after an action took place.
        
        Parameters
        ----------
        new_display_state : ``DisplayState``
            The new display state to render.
        """
        old_display_state = self.display_state
        self.display_state = new_display_state
        
        prefix_length = self.prefix_length
        content_width = get_new_content_width(prefix_length)
        
        write_buffer = []
        
        if old_display_state.content_width == content_width:
            new_display_state.write_difference(
                old_display_state, write_buffer, prefix_length, self.prefix_initial, self.prefix_continuous
            )
        else:
            old_display_state.clear(write_buffer, prefix_length, content_width)
            self.display_state.write(
                write_buffer, prefix_length, self.prefix_initial, self.prefix_continuous, content_width
            )
            new_display_state.write_cursor(write_buffer, prefix_length)
        
        try:
            self._write_write_buffer(write_buffer)
        finally:
            write_buffer = None
    
    
    def _process_async_content(self, data):
        """
        Processes the given async content and decides whether it should be flushed before the editor's input.
        
        Parameters
        ----------
        data : `bytes`
            Raw content to write.
        """
        async_output_data_continuous = self.async_output_data_continuous
        if (async_output_data_continuous is not None):
            data = async_output_data_continuous + data
            self.async_output_data_continuous = None
        
        content, self.async_output_data_continuous = try_decode(data)
        
        async_output_content_continuous = self.async_output_content_continuous
        if (async_output_content_continuous is not None):
            content = async_output_content_continuous + content
            self.async_output_content_continuous = None
        
        if not content.endswith('\n'):
            line_break_index = content.rfind('\n')
            if line_break_index == -1:
                self.async_output_content_continuous = content
                return
            
            line_break_index += 1
            self.async_output_content_continuous = content[line_break_index:]
            content = content[:line_break_index]
        
        if content:
            self._display_async_content(content)
    
    
    def _maybe_display_continuous_content(self):
        """
        If there is async content to display, displays it.
        """
        content = None
        
        async_output_data_continuous = self.async_output_data_continuous
        if (async_output_data_continuous is not None):
            content = async_output_data_continuous.decode(errors = 'backslashreplace')
            self.async_output_data_continuous = None
        
        async_output_content_continuous = self.async_output_content_continuous
        if (async_output_content_continuous is not None):
            if content is None:
                content = async_output_content_continuous
            else:
                content = async_output_content_continuous + content
        
        if (content is not None):
            self._display_async_content(content)
    
    
    def _display_async_content(self, content):
        """
        Writes the given content before the editor's input.
        
        Parameters
        ----------
        content : `str`
            The content to write.
        """
        self.output_stream.flush()
        
        # Clear display state
        write_buffer = []
        
        async_output_written = self.async_output_written
        prefix_length = self.prefix_length
        
        content_width = get_new_content_width(prefix_length)
        
        self.display_state.clear(write_buffer, prefix_length, content_width)
        
        if async_output_written:
            write_buffer.append(COMMAND_UP)
        
        try:
            self._write_write_buffer(write_buffer)
        finally:
            write_buffer = None
        write_buffer = []
        
        # Write async content
        termios.tcsetattr(self.input_stream, termios.TCSADRAIN, self.input_stream_settings_original)
        write_buffer.append(content)
        
        try:
            self._write_write_buffer(write_buffer)
        finally:
            write_buffer = None
        write_buffer = []
        
        tty.setraw(self.input_stream)
        
        # Rewrite display state
        write_buffer.append(COMMAND_DOWN)
        write_buffer.append(COMMAND_START_LINE)
        
        if not async_output_written:
            self.async_output_written = True
        
        self.display_state.write(
            write_buffer, prefix_length, self.prefix_initial, self.prefix_continuous, content_width
        )
        self.display_state.write_cursor(write_buffer, prefix_length)
        
        try:
            self._write_write_buffer(write_buffer)
        finally:
            write_buffer = None
    
    
    def check_exit_conditions(self, old_display_state):
        """
        Checks whether any exit conditions are met.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Raises
        ------
        SyntaxError
        """
        if _check_should_try_compile(old_display_state):
            compiled_code = maybe_compile(old_display_state.buffer, self.file_name)
            if (compiled_code is not None):
                self.compiled_code = compiled_code
                self.alive = False
                return True
        
        return False
    
    
    def should_move_in_history(self):
        """
        Returns whether history moving can be executed.
        
        Returns
        -------
        should_move_in_history : `bool`
        """
        display_state = self.display_state
        if display_state.cursor_line_index != 0:
            return False
        
        display_state = self.display_state
        if display_state.cursor_index != 0:
            return False
        
        return self.can_move_in_history()
        
    
    def can_move_in_history(self):
        """
        Returns whether history moving can be executed.
        
        Returns
        -------
        can_move_in_history : `bool`
        """
        if not self.modified:
            return True
        
        if self.is_empty():
            self.modified = False
            return True
        
        return False
    
    
    def try_rewind_in_history(self):
        """
        Tries to go back in history if applicable.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if not self.should_move_in_history():
            return None
        
        buffer = self.history.get_previous()
        if buffer is None:
            buffer = ['']
        
        display_state = self.display_state.copy()
        display_state.buffer = buffer.copy()
        return display_state
    
    
    def try_forward_in_history(self):
        """
        Tries to go forward in history if applicable.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if not self.should_move_in_history():
            return None
        
        buffer = self.history.get_next()
        if buffer is None:
            buffer = ['']
        
        display_state = self.display_state.copy()
        display_state.buffer = buffer.copy()
        return display_state
    
    
    def checkout_indexed_input(self, old_display_state):
        """
        Checks out whether an index of the history. If it is returns a new display state.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        index = -1
        
        for line in old_display_state.buffer:
            if _check_is_line_empty(line):
                continue
            
            if index != -1:
                return None
            
            matched = INDEXED_INPUT_RP.fullmatch(line)
            if matched is None:
                return None
            
            index = int(matched.group(1))
            continue
        
        if index == -1:
            return None
        
        buffer = self.history.get_at(index)
        return DisplayState(buffer)
    
    
    def _write_write_buffer(self, write_buffer):
        if not write_buffer:
            return
        
        try:
            data = memoryview(''.join(write_buffer).encode())
        finally:
            write_buffer = None
        
        file_descriptor = self.output_stream.fileno()
        if not get_blocking(file_descriptor):
            set_blocking(file_descriptor, False)
        
        while True:
            # Try to write now.
            try:
                n = write(file_descriptor, data)
            except (BlockingIOError, InterruptedError):
                pass
            else:
                data = data[n:]
                if not data:
                    return
            
            # Wait till io becomes writable again, or timeout.
            poller = Poller()
            poller.register(file_descriptor, EVENT_POLL_WRITE)
            
            now = monotonic()
            poll_end = now + WRITE_TIMEOUT
            
            while True:
                events = poller.poll(poll_end - now)
                if events:
                    break
                
                now = monotonic()
                if now < poll_end:
                    continue
                
                raise RuntimeError(
                    f'The file descriptor did not became writable within timeout ({WRITE_TIMEOUT}).'
                )



__all__ = ('ExceptionRepresentationGeneric',)

from ...docs import copy_docs

from .exception_representation_base import ExceptionRepresentationBase
from .representation_helpers import get_exception_representation_generic


class ExceptionRepresentationGeneric(ExceptionRepresentationBase):
    """
    Holds a generic exception's representation.
    
    Attributes
    ----------
    representation : `str`
        The exception's representation.
    """
    __slots__ = ('representation',)
    
    def __new__(cls, exception, frames):
        """
        Creates a new exception representation.
        
        Parameters
        ----------
        exception : `BaseException`
            Exception to represent.
        frames : `None | list<FrameProxyBase>`
            The frames the exception is raised with.
        """
        representation = get_exception_representation_generic(exception)
        
        self = object.__new__(cls)
        self.representation = representation
        return self
    
    
    @classmethod
    def from_fields(cls, *, representation = ...):
        """
        Creates a new generic exception representation from the given fields.
        
        Parameters
        ----------
        representation : `str`
            The exception's representation.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.representation = '' if representation is ... else representation
        return self
    
    
    @copy_docs(ExceptionRepresentationBase._populate_repr_parts)
    def _populate_repr_parts(self, repr_parts):
        repr_parts.append(' representation = ')
        repr_parts.append(repr(self.representation))
    
    
    @copy_docs(ExceptionRepresentationBase._is_equal)
    def _is_equal(self, other):
        if self.representation != other.representation:
            return False
        
        return True



__all__ = ('get_exception_representation',)

from .attribute_error_helpers import is_attribute_error
from .exception_helpers import is_exception
from .exception_representation_attribute_error import ExceptionRepresentationAttributeError
from .exception_representation_generic import ExceptionRepresentationGeneric
from .exception_representation_syntax_error import ExceptionRepresentationSyntaxError
from .syntax_error_helpers import is_syntax_error



SELECTORS = [
    (is_attribute_error, ExceptionRepresentationAttributeError),
    (is_syntax_error, ExceptionRepresentationSyntaxError),
    (is_exception, ExceptionRepresentationGeneric),
]


def get_exception_representation(exception, frames):
    """
    Gets the given exception's representation.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to get its representation of.
    frame : `None | list<FrameProxyBase>`
        The frames the exception is raised with.
    
    Returns
    -------
    exception_representation : `None | ExceptionRepresentationBase`
    """
    for checker, representation_type in SELECTORS:
        if checker(exception):
            return representation_type(exception, frames)
    
    return None



from .exception_proxy_base import *
from .exception_proxy_rich import *
from .exception_proxy_virtual import *


__all__ = (
    *exception_proxy_base.__all__,
    *exception_proxy_rich.__all__,
    *exception_proxy_virtual.__all__,
)



__all__ = ('alchemy_incendiary', 'call', 'copy_func',)

from types import FunctionType

from .docs import has_docs


@has_docs
def call(function):
    """
    Calls the function returning itself.
    
    Parameters
    ----------
    function : `object`
        The function to call
    
    Returns
    -------
    function : `object`
        The function itself.
    """
    function()
    return function


@has_docs
class alchemy_incendiary:
    """
    Function wrapper familiar to `functools.partial`.
    
    Used by hata to run functions inside of executors.
    
    Attributes
    ----------
    positional_parameters : `tuple` of `object`
        Parameters to call `func` with.
    function : `callable`
        The function to call.
    keyword_parameters : `None` of `dict` of (`str`, `object`) items
        Keyword parameters to call func with if applicable.
    """
    __slots__ = ('positional_parameters', 'function', 'keyword_parameters',)
    
    
    @has_docs
    def __init__(self, function, positional_parameters, keyword_parameters = None):
        """
        Creates a new `alchemy_incendiary` with the given parameters.
        
        Parameters
        ----------
        function : `callable`
            The function to call.
        positional_parameters : `tuple<object>`
            Parameters to call `func` with.
        keyword_parameters : `None | dict<str, object>` = `None`, Optional
            Keyword parameters to call func with if applicable.
        """
        self.function = function
        self.positional_parameters = positional_parameters
        self.keyword_parameters = keyword_parameters
    
    
    @has_docs
    def __call__(self):
        """
        Calls the ``alchemy_incendiary``'s inner function with t's parameters and keyword parameters.
        
        Returns
        -------
        result : `object`
            The returned value by ``.func``.
        
        Raises
        ------
        BaseException
            The raised exception by ``.func``.
        """
        keyword_parameters = self.keyword_parameters
        if keyword_parameters is None:
            return self.function(*self.positional_parameters)
        
        return self.function(*self.positional_parameters, **keyword_parameters)


@has_docs
def copy_func(old):
    """
    Copies the given function.
    
    Parameters
    ----------
    old : `function`
        The function to copy.
    
    Returns
    -------
    new : `functions`
        The new created function.
    """
    new = FunctionType(
        old.__code__,
        old.__globals__,
        name = old.__name__,
        argdefs = old.__defaults__,
        closure = old.__closure__,
    )
    new.__kwdefaults__ = old.__kwdefaults__
    return new



__all__ = ('ClientRequest',)

from http.cookies import Morsel, SimpleCookie
from warnings import warn

from ..core import CancelledError, Task
from ..utils import IgnoreCaseMultiValueDictionary, RichAttributeErrorBaseType
from ..web_common.form_data import FormData
from ..web_common.headers import (
    AUTHORIZATION, CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TYPE, COOKIE, HOST, METHOD_CONNECT, PROXY_AUTHORIZATION,
    TRANSFER_ENCODING
)
from ..web_common import BasicAuthorization
from ..web_common.http_stream_writer import HTTPStreamWriter
from ..web_common.multipart import create_payload

from .client_response import ClientResponse
from .connection_key import ConnectionKey
from .constants import DEFAULT_HEADERS
from .request_info import RequestInfo


class ClientRequest(RichAttributeErrorBaseType):
    """
    Http request class used by ``HTTPClient``.
    
    Attributes
    ----------
    authorization : `None | BasicAuthorization`
        Authorization sent with the request.
    
    body : `None | PayloadBase`
        The request's body.
    
    chunked : `bool`
        Whether the request is sent chunked.
    
    compression : `None | str`
        Compression used when sending the request.
    
    headers : ``IgnoreCaseMultiValueDictionary``
        The headers of the request.
    
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    
    method : `str`
        The request's method.
    
    original_url : ``URL``
        The original url, what was asked to request.
    
    proxied_url : `None | Url`
        The proxied url that the requests connects to.
    
    proxy : `None | Proxy`
        Proxy if applicable.4
    
    response : `None | ClientResponse`
        Object representing the received response. Set as `None` till ``.send`` finishes.
    
    ssl_context : `None | SSLContext`
        The connection's ssl type.
    
    ssl_fingerprint : `None | SSLFingerprint`
        Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
    
    url : ``URL``
        The url, what will be requested.
    
    write_body_task : `None | Task<.write_body>`
        Payload writer task, what is present meanwhile the request's payload is sending.
    """
    __slots__ = (
        'authorization', 'body', 'chunked', 'compression', 'headers', 'loop', 'method', 'original_url', 'proxied_url',
        'proxy', 'response', 'ssl_context', 'ssl_fingerprint', 'url', 'write_body_task'
    )
    
    def __new__(
        cls,
        loop,
        method,
        url,
        headers,
        data,
        query,
        cookies,
        authorization,
        proxied_url,
        proxy,
        ssl_context,
        ssl_fingerprint,
    ):
        """
        Creates a new client request with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop, trough what the request is executed.
        
        method : `str`
            The request's method.
        
        url : ``URL``
            The url to request.
        
        headers : ``IgnoreCaseMultiValueDictionary``
            Headers of the request.
        
        data : `None | bytes-like | io-like | FormData`
            Data to send as the request's body.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`
            Query string parameters.
        
        cookies : `None | dict<str, str | Morsel>`
            Cookies OwO.
        
        authorization : `None | BasicAuthorization`
            Authorization sent with the request.
        
        proxy : `None | Proxy`
            Proxy if applicable.
        
        ssl_context : `None | SSLContext``
            The connection's ssl type.
        
        ssl_fingerprint : `None | SSLFingerprint`
            Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
        
        Raises
        ------
        TypeError
            - ˙Cannot serialize a field of the given `data`.
        ValueError
            - Host could not be detected from `url`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        """
        # check url
        if url.host is None:
            raise ValueError(f'Host could not be detected from url: {url!r}')
        
        # check proxied_url
        if (method == METHOD_CONNECT):
            if proxied_url is None:
                raise TypeError(f'`proxied_url` cannot be `None` if `method == METHOD_CONNECT`.')
            
            if proxied_url.host is None:
                raise ValueError(f'Host could not be detected from url: {proxied_url!r}')
        
        # Add extra query parameters to the url and remove fragments
        url = url.extend_query(query)
        request_url = url.with_fragment(None)
        
        # Check authorization
        if authorization is None:
            # If authorization is given, try to detect from url.
            user_id = url.user
            password = url.password
            
            if (user_id is not None) and user_id:
                if password is None:
                    password = ''
                
                authorization = BasicAuthorization(user_id, password)
        
        # Store authorization in headers is applicable.
        # Use `proxy_authorization` header if we are doing a connect request. 
        if (authorization is not None):
            headers[PROXY_AUTHORIZATION if method == METHOD_CONNECT else AUTHORIZATION] = authorization.to_header()
        
        # Add default headers. Except if we are making a connect request.
        if (method != METHOD_CONNECT):
            for key, value in DEFAULT_HEADERS:
                headers.setdefault(key, value)
        
        # Add host to headers if not present. Except if we are making a connect request.
        if (method != METHOD_CONNECT) and (HOST not in headers):
            location = request_url.raw_host.rstrip('.')
            if not request_url.is_default_port():
                port = request_url.port
                if (port is not None):
                    location = f'{location}:{port}'
            
            headers[HOST] = location
        
        # Update cookies
        if (cookies is not None) and cookies:
            cookie = SimpleCookie()
            
            cookie_header = headers.pop(COOKIE, '')
            if cookie_header:
                cookie.load(cookie_header)
            
            for key, value in cookies.items():
                if isinstance(key, Morsel):
                    try:
                        morsel_value = value[value.key]
                    except KeyError:
                        morsel_value = Morsel()
                    
                    morsel_value.set(value.key, value.value, value.coded_value)
                    value = morsel_value
                
                cookie[key] = value
            
            headers[COOKIE] = cookie.output(header = '', sep = ';').strip()
        
        # analyze headers & data
        if (data is not None) and (not data):
            data = None
        
        if (data is None):
            chunked = False
            compression = None
            
            # Remove not needed headers when not transferring any data.
            for header_name in (CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TYPE, TRANSFER_ENCODING):
                try:
                    del headers[header_name]
                except KeyError:
                    pass
        
        else:
            # Needed for transfer data checks
            chunked = 'chunked' in headers.get(TRANSFER_ENCODING, '').casefold()
            compression = headers.get(CONTENT_ENCODING, None)
            
            if (compression is not None):
                chunked = True
            
            # form_data
            if isinstance(data, FormData):
                data = data.generate_form()
            else:
                try:
                    data = create_payload(data, {'disposition': None})
                except LookupError:
                    data = FormData.from_fields(data).generate_form()
            
            if not chunked:
                if CONTENT_LENGTH not in headers:
                    size = data.size
                    if size is None:
                        chunked = True
                    else:
                        if CONTENT_LENGTH not in headers:
                            headers[CONTENT_LENGTH] = str(size)
            
            if CONTENT_TYPE not in headers:
                headers[CONTENT_TYPE] = data.content_type
            
            data_headers = data.headers
            if data_headers:
                for key, value in data_headers.items():
                    headers.setdefault(key, value)
            
            # Analyze transfer-encoding header.
            if chunked:
                if CONTENT_LENGTH in headers:
                    raise ValueError(
                        f'Chunked is mutually exclusive with the `{CONTENT_LENGTH!s}` header.'
                    )
                
                # We may have modify `chunked` value since the start, so lets set it again to make sure.
                headers[TRANSFER_ENCODING] = 'chunked'
            
            else:
                if CONTENT_LENGTH not in headers:
                    headers[CONTENT_LENGTH] = '0' if data is None else str(len(data))
        
        # Everything seems correct, create the object.
        self = object.__new__(cls)
        self.authorization = authorization
        self.body = data
        self.chunked = chunked
        self.compression = compression
        self.headers = headers
        self.loop = loop
        self.method = method
        self.original_url = url
        self.proxied_url = proxied_url
        self.proxy = proxy
        self.response = None
        self.ssl_context = ssl_context
        self.ssl_fingerprint = ssl_fingerprint 
        self.url = request_url
        self.write_body_task = None
        
        return self
    
    
    def __repr__(self):
        """Returns the client requests' representation."""
        repr_parts = ['<', type(self).__name__]
        
        # method
        repr_parts.append(' method = ')
        repr_parts.append(repr(self.method))
        
        # url
        repr_parts.append(', url = ')
        repr_parts.append(repr(self.url))
        
        # headers
        repr_parts.append(', headers = ')
        repr_parts.append(repr(self.headers))
        
        # original_url
        original_url = self.original_url
        if self.url != original_url:
            repr_parts.append(', original_url = ')
            repr_parts.append(repr(original_url))
        
        # authorization
        authorization = self.authorization
        if (authorization is not None):
            repr_parts.append(', authorization = ')
            repr_parts.append(repr(authorization))
        
        # body
        body = self.body
        if (body is not None):
            repr_parts.append(', type(body) = ')
            repr_parts.append(type(body).__name__)
        
        # chunked
        chunked = self.chunked
        if chunked:
            repr_parts.append(', chunked = ')
            repr_parts.append(repr(chunked))
        
        # compression
        compression = self.compression
        if (compression is not None):
            repr_parts.append(', compression = ')
            repr_parts.append(repr(compression))
        
        # proxied_url
        proxied_url = self.proxied_url
        if (proxied_url is not None):
            repr_parts.append(', proxied_url = ')
            repr_parts.append(repr(proxied_url))
        
        # proxy
        proxy = self.proxy
        if (proxy is not None):
            repr_parts.append(', proxy = ')
            repr_parts.append(repr(proxy))
        
        # ssl_context
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            repr_parts.append(', ssl_context = ')
            repr_parts.append(repr(ssl_context))
        
        # ssl_fingerprint
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            repr_parts.append(', ssl_fingerprint = ')
            repr_parts.append(repr(ssl_fingerprint))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def is_secure(self):
        """
        Returns whether the request is secure.
        
        Returns
        -------
        is_secure : `bool`
        """
        scheme = self.url.scheme
        if scheme in ('http', 'ws'):
            return False
        
        if scheme in ('https', 'wss'):
            return True
        
        # If we have any other scheme let the user decide whether it is ssl or nah.
        return (self.ssl_context is not None)
    
    
    @property
    def connection_key(self):
        """
        Returns the connection key of request.
        
        Returns
        -------
        connection_key : ``ConnectionKey``
        """
        return ConnectionKey(
            self.host,
            self.port,
            self.proxy,
            self.is_secure(),
            self.ssl_context,
            self.ssl_fingerprint,
        )
    
    
    @property
    def request_info(self):
        """
        Returns base information representing the request.
        
        Returns
        -------
        request_info : ``RequestInfo``
        """
        return RequestInfo(
            self.headers,
            self.method,
            self.original_url,
            self.url,
        )
    
    
    @property
    def host(self):
        """
        Returns the request's host.
        
        Returns
        -------
        host : `str`
        """
        return self.url.host
    
    
    @property
    def port(self):
        """
        Returns the request's port.
        
        Returns
        -------
        port : `None | int`
        """
        return self.url.port
    
    
    async def write_body(self, connection):
        """
        Writes the request's body.
        
        This method is a coroutine.
        
        Parameters
        ----------
        connection : ``Connection``
            Connection of the request with what the payload is sent.
        """
        writer = HTTPStreamWriter(connection.protocol, self.compression, self.chunked)
        
        try:
            body = self.body
            if (body is not None):
                await self.body.write(writer)
        except OSError as exception:
            new_exception = OSError(exception.errno, f'Failed to write request body for {self.url!r}.')
            new_exception.__cause__ = exception
            connection.protocol.set_exception(new_exception)
        
        except GeneratorExit:
            raise
        
        except CancelledError:
            await writer.write_eof()
            raise
        
        except BaseException as exception:
            connection.protocol.set_exception(exception)
            raise
        
        else:
            await writer.write_eof()
        
        finally:
            self.write_body_task = None
    
    
    def begin(self, connection):
        """
        Begins sending the request.
        
        Parameters
        ----------
        connection : ``Connection``
            Connection, what is used to send the request.
        
        Returns
        -------
        response : ``ClientResponse``
        """
        try:
            # Note: perhaps move `write_http_request` to `HTTPStreamWriter`. 
            connection.protocol.write_http_request(self.method, self._build_path_to_request(), self.headers)
            self.write_body_task = Task(self.loop, self.write_body(connection))
            
            self.response = response = ClientResponse(self, connection)
            
            return response
        
        except:
            connection.close()
            raise
    
    
    def _build_path_to_request(self):
        """
        Builds the path to be requested.
        
        Returns
        -------
        path : `str`
        """
        if self.method == METHOD_CONNECT:
            url = self.proxied_url
            path = url.raw_host
            port = url.port
            if (port is not None):
                path = f'{path}:{port}'
        
        else:
            url = self.url
            path = url.raw_path
            raw_query_string = url.raw_query_string
            if (raw_query_string is not None):
                path = f'{path}?{raw_query_string}'
        
        return path
    
    
    @property
    def auth(self):
        """
        ``.auth`` is deprecated and will be removed in 2025 November.
        Please use ``.authorization`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.auth` is deprecated and will be removed in 2025 November. '
                f'Please use `.authorization instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.authorization

    
    @property
    def proxy_auth(self):
        """
        ``.proxy_auth`` is deprecated and will be removed in 2025 November.
        Please use ``.proxy.authorization`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.proxy_auth` is deprecated and will be removed in 2025 November. '
                f'Please use `.proxy.authorization instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        proxy = self.proxy
        if (proxy is not None):
            return proxy.authorization

    
    @property
    def proxy_headers(self):
        """
        ``.proxy_headers`` is deprecated and will be removed in 2025 November.
        Please use ``.proxy.headers`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.proxy_headers` is deprecated and will be removed in 2025 November. '
                f'Please use `.proxy.headers instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        proxy = self.proxy
        if (proxy is not None):
            return proxy.headers
    
    
    @property
    def proxy_url(self):
        """
        ``.proxy_url`` is deprecated and will be removed in 2025 November.
        Please use ``.proxy.url`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.proxy_url` is deprecated and will be removed in 2025 November. '
                f'Please use `.proxy.url instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        proxy = self.proxy
        if (proxy is not None):
            return proxy.url
